from enum import Enum
from typing import Callable, Union, List, Optional

from numpy.typing import NDArray
from shapely import Polygon, MultiPolygon

from .contour import contour_inward, contour_outward
from .raster import (horizontal_left_to_right, horizontal_right_to_left,
                     vertical_bottom_to_top, vertical_top_to_bottom)
from .serpentine import serpentine_vertical, serpentine_horizontal
from ..beams import BeamType
from ..mill import Mill

_method_signature = Callable[[Union[Polygon, MultiPolygon], Mill, BeamType, Optional[bool]], List[NDArray]]
_method_map = {
    1: vertical_top_to_bottom,
    2: vertical_bottom_to_top,
    3: horizontal_left_to_right,
    4: horizontal_right_to_left,
    5: serpentine_vertical,
    6: serpentine_horizontal,
    7: contour_outward,
    8: contour_inward
}


class ScanStyle(Enum):

    @property
    def value(self) -> _method_signature:
        """Override value to return the corresponding function instead of the integer."""
        return _method_map[super().value]


class ScanStyles:

    class Sequential(ScanStyle):
        vertical_top_to_bottom = 1
        vertical_bottom_to_top = 2
        horizontal_left_to_right = 3
        horizontal_right_to_left = 4

    class Serpentine(ScanStyle):
        vertical = 5
        horizontal = 6

    class Contour(ScanStyle):
        outwards = 7
        inwards = 8



from __future__ import annotations

from abc import ABC
from typing import ClassVar

if __name__ != "__main__":
    from ..resources import length_units, LENGTHUNITS

_volume_pr_doses = {
    "C": 0.18,
    "Default_Si": 0.15,
    "PZT": 0.24,
    "Si": 0.27,
    "Al": 0.30,
    "Ti": 0.37,
    "Cr": 0.10,
    "Fe": 0.29,
    "Ni": 0.14,
    "Cu": 0.25,
    "Mo": 0.12,
    "Ta": 0.32,
    "W": 0.12,
    "Au": 1.50,
    "MgO": 0.15,
    "SiO2": 0.24,
    "Al2O3": 0.08,
    "TiO": 0.15,
    "Si3N4": 0.20,
    "TiN": 0.15,
    "Fe2O3": 0.25,
    "GaAs": 0.20,
    "Pt": 0.23,
    "PMMA": 0.40
}


# Print all material classes for copy pasting if you change or add values in the volume pr. doses dict
def _print_classes():
    string = "class Materials:\n    __slots__ = ()\n\n"
    for k, v in _volume_pr_doses.items():
        string += f"    class {k}(Material):\n        _volume_pr_dose = _volume_pr_doses['{k}']\n\n"
    print(string)


class Material(ABC):
    _volume_pr_dose: ClassVar[float]
    _thickness: float

    __slots__ = ["_volume_pr_dose", "_thickness"]

    def __init__(self, thickness: float = float('inf'), units: LENGTHUNITS = "um"):
        if thickness == float('inf'):
            self._thickness = thickness
        else:
            self._thickness = length_units(thickness, units, "m")

    def get_thickness(self, units: LENGTHUNITS = "um") -> float:
        """Returns the thickness of the material in the specified units."""
        if self._thickness == float('inf'):
            return float('inf')
        else:
            return length_units(self._thickness, "m", units)

    @property
    def volume_pr_dose(self) -> float:
        """Returns the volume pr. dose for the material in units of microns cubed over nanocoloumbs [um^3/nC]."""
        return self._volume_pr_dose

    def __repr__(self) -> str:
        thickness = "inf" if self._thickness == float('inf') else f"{self.get_thickness('um')} μm"
        return (f"{self.__class__.__name__}(thickness={thickness}, "
                f"volume_pr_dose={self._volume_pr_dose} μm³/nC)")


class Materials:
    __slots__ = ()

    class C(Material):
        _volume_pr_dose = _volume_pr_doses['C']

    class Default_Si(Material):
        _volume_pr_dose = _volume_pr_doses['Default_Si']

    class PZT(Material):
        _volume_pr_dose = _volume_pr_doses['PZT']

    class Si(Material):
        _volume_pr_dose = _volume_pr_doses['Si']

    class Al(Material):
        _volume_pr_dose = _volume_pr_doses['Al']

    class Ti(Material):
        _volume_pr_dose = _volume_pr_doses['Ti']

    class Cr(Material):
        _volume_pr_dose = _volume_pr_doses['Cr']

    class Fe(Material):
        _volume_pr_dose = _volume_pr_doses['Fe']

    class Ni(Material):
        _volume_pr_dose = _volume_pr_doses['Ni']

    class Cu(Material):
        _volume_pr_dose = _volume_pr_doses['Cu']

    class Mo(Material):
        _volume_pr_dose = _volume_pr_doses['Mo']

    class Ta(Material):
        _volume_pr_dose = _volume_pr_doses['Ta']

    class W(Material):
        _volume_pr_dose = _volume_pr_doses['W']

    class Au(Material):
        _volume_pr_dose = _volume_pr_doses['Au']

    class MgO(Material):
        _volume_pr_dose = _volume_pr_doses['MgO']

    class SiO2(Material):
        _volume_pr_dose = _volume_pr_doses['SiO2']

    class Al2O3(Material):
        _volume_pr_dose = _volume_pr_doses['Al2O3']

    class TiO(Material):
        _volume_pr_dose = _volume_pr_doses['TiO']

    class Si3N4(Material):
        _volume_pr_dose = _volume_pr_doses['Si3N4']

    class TiN(Material):
        _volume_pr_dose = _volume_pr_doses['TiN']

    class Fe2O3(Material):
        _volume_pr_dose = _volume_pr_doses['Fe2O3']

    class GaAs(Material):
        _volume_pr_dose = _volume_pr_doses['GaAs']

    class Pt(Material):
        _volume_pr_dose = _volume_pr_doses['Pt']

    class PMMA(Material):
        _volume_pr_dose = _volume_pr_doses['PMMA']

if __name__ == "__main__":
    _print_classes()


import sys

from . import shapes
from .scan_styles import ScanStyles
from .fib import FIBG2
from .resources import plot_stream_file, print_streamfile_time

REQUIRED_MAJOR = 3
REQUIRED_MINOR = 11  # Change this to your minimum required version

if sys.version_info < (REQUIRED_MAJOR, REQUIRED_MINOR):
    raise RuntimeError(
        f"The deFIBrillator package requires Python {REQUIRED_MAJOR}.{REQUIRED_MINOR} or higher."
        f"You are currently using Python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}."
    )

__all__ = ["shapes", "FIBG2", "plot_stream_file", "print_streamfile_time", "ScanStyles"]



from typing import List, Union
import itertools

import numpy as np
from numpy.typing import NDArray
from shapely import Point, LineString, MultiLineString, Polygon, MultiPolygon, buffer, prepare, intersection

from .raster import _raster_sequential
from ..beams import BeamType
from ..mill import Mill


def serpentine_horizontal(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                          horizontal: bool = True) -> List[NDArray]:
    streamlines = _raster_sequential(shape, mill, beam, horizontal=True)

    tolerance = beam.get_radius("nm")
    streamlines.sort(key=lambda s: (int(s[0, 2] / tolerance), s[0, 1]))

    # Group the streamlines by the bucketed y coordinate.
    grouped_streamlines = []
    for group_key, group in itertools.groupby(streamlines, key=lambda s: int(s[0, 2] / tolerance)):
        grouped_streamlines.append(list(group))

    # Now, for every other pass (group), reverse the order of points in each scan line.
    final_streamlines = []
    for i, group in enumerate(grouped_streamlines):
        if i % 2 == 1:  # odd-indexed pass: reverse the scan direction
            group = [line[::-1] for line in group][::-1]
        final_streamlines.extend(group)

    return final_streamlines


def serpentine_vertical(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                        horizontal: bool = True) -> List[NDArray]:

    streamlines = _raster_sequential(shape, mill, beam, horizontal=False)

    tolerance = beam.get_radius("nm")
    streamlines.sort(key=lambda s: (int(s[0, 1] / tolerance), s[0, 2]))

    # Group the streamlines by the bucketed y coordinate.
    grouped_streamlines = []
    for group_key, group in itertools.groupby(streamlines, key=lambda s: int(s[0, 1] / tolerance)):
        grouped_streamlines.append(list(group))

    # Now, for every other pass (group), reverse the order of points in each scan line.
    final_streamlines = []
    for i, group in enumerate(grouped_streamlines):
        if i % 2 == 1:  # odd-indexed pass: reverse the scan direction
            group = [line[::-1] for line in group][::-1]
        final_streamlines.extend(group)

    return final_streamlines



from __future__ import annotations

# Std lib
from typing import Unpack, TypedDict, Literal

# 3rd party lib
from matplotlib.font_manager import FontProperties
from matplotlib.textpath import TextPath
from shapely import Polygon, MultiPolygon, unary_union
from shapely.affinity import translate

# Local lib
from .shape import Shape


class TextKwargs(TypedDict, total=False):
    x: float
    y: float
    font_height: float
    font: Literal[
        # Windows Fonts
        "Arial", "Arial Black", "Calibri", "Cambria", "Candara", "Comic Sans MS",
        "Consolas", "Courier New", "Georgia", "Impact", "Lucida Console",
        "Lucida Sans Unicode", "Microsoft Sans Serif", "Segoe UI", "Tahoma",
        "Times New Roman", "Trebuchet MS", "Verdana",

        # macOS Fonts
        "American Typewriter", "Arial", "Avenir", "Big Caslon", "Brush Script",
        "Chalkboard", "Courier", "Geneva", "Gill Sans", "Helvetica",
        "Helvetica Neue", "Hoefler Text", "Menlo", "Monaco", "Optima",
        "Palatino", "Snell Roundhand", "Times", "Zapfino",

        # Linux Fonts (Commonly Available)
        "DejaVu Sans", "DejaVu Serif", "Liberation Sans", "Liberation Serif",
        "Liberation Mono", "Noto Sans", "Noto Serif", "Noto Mono", "Ubuntu",
        "Ubuntu Mono"
    ]
    boldness: int
    millable: bool
    millable_override: bool


class Text(Shape):

    # region Class Body

    # Class variables
    _default_geometry = Shape._default_geometry.copy()
    _default_geometry.update({"font_height": 1000, "font": "arial", "boldness": 500})

    # Instance variables
    _font_height: float
    _font: str
    _boldness: int
    _text: str

    __slots__ = ["_font_height", "_text", "_font", "_boldness"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, text: str, **kwargs: Unpack[TextKwargs]) -> None:
        super().__init__(**kwargs)
        self._text = text

    def _get_polygon(self) -> MultiPolygon:
        """
        Converts text to a Shapely MultiPolygon while preserving holes,
        with support for bold fonts.
        """

        # Set up font properties (change family and style as needed)
        try:
            font_prop = FontProperties(family=self._font, weight=self._boldness)  # Example: Arial Bold
        except ValueError as e:
            if f"weight={self._boldness} is invalid" in str(e):
                raise ValueError(f"The boldness of a font can only be between 0 and 1000, got {self._boldness}.")
            else:
                raise e

        # Create the text path
        abspos = self._get_absolute_position()
        path = TextPath((float(abspos[0]), float(abspos[1])), self._text, size=self._font_height * (1 / 0.725),
                        prop=font_prop)

        # Process the polygons (same as before)
        contours = [Polygon(vertices) for vertices in path.to_polygons() if len(vertices) >= 3]

        if not contours:
            return MultiPolygon()

        contours.sort(key=lambda p: p.area, reverse=True)
        polygons = []
        used_as_hole = set()

        for i, outer in enumerate(contours):
            if i in used_as_hole:
                continue
            holes = []
            for j, hole in enumerate(contours):
                if j != i and j not in used_as_hole and outer.contains(hole):
                    holes.append(hole.exterior)
                    used_as_hole.add(j)
            polygons.append(Polygon(outer.exterior, holes))

        final_geometry = MultiPolygon(polygons)
        minx, miny, maxx, maxy = final_geometry.bounds
        centered_geometry = translate(final_geometry, xoff=-(maxx - minx) / 2, yoff=-(maxy - miny) / 2)

        return centered_geometry

    # endregion Dev Methods

    # region User Methods

    def copy(self, **kwargs) -> Text:
        return super().copy(**kwargs)

    # endregion User Methods

    # region Properties

    @property
    def x_span(self) -> float:
        return self.x_max - self.x_min

    @property
    def y_span(self) -> float:
        return self.y_max - self.y_min

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, text: str):
        self._text = str(text)

    @property
    def font_size(self) -> int:
        return self._font_height

    @font_size.setter
    def font_size(self, value: int):
        self._font_height = int(value)

    # endregion Properties



# 3rd party lib
import sys

import numpy as np
from shapely import LineString


def find_corner_points(coords, angle_threshold=30):
    """Find indices of corner points where the angle is below the threshold."""
    dx = np.diff(coords[:, 0])
    dy = np.diff(coords[:, 1])

    angles = np.arctan2(dy, dx)

    # Compute absolute angle differences in-place
    angle_diffs = np.abs(angles[1:] - angles[:-1]) * 180 / np.pi
    angle_diffs = np.minimum(angle_diffs, 360 - angle_diffs)  # Ensure within [0, 180]

    # Find indices where the angle is below the threshold
    corner_indices = np.flatnonzero(angle_diffs > angle_threshold) + 1  # +1 shifts indices forward

    # Always include first and last points
    return np.r_[0, corner_indices, len(coords) - 1]  # `np.r_` avoids explicit `concatenate`


def interpolate_segment_with_original_points(segment_coords: np.ndarray, step_size: float) -> np.ndarray:
    """Interpolate a single segment while keeping original points."""
    segment = LineString(segment_coords)

    # Compute cumulative distances along the segment
    dxdy = np.diff(segment_coords, axis=0)
    seg_lengths = np.hypot(dxdy[:, 0], dxdy[:, 1])
    cum_lengths = np.r_[0, np.cumsum(seg_lengths)]  # `np.r_` for efficiency
    total_length = cum_lengths[-1]

    # Ensure original points are included
    num_steps = max(1, int(np.rint(total_length / step_size)))
    distances = np.linspace(0, total_length, num_steps)

    return interpolate_line(segment, distances)


def interpolate_curve(line: LineString, step_size: float, angle_threshold=30):
    """Segment the LineString at corners and interpolate each segment while keeping original points."""
    coords = np.array(line.coords)
    corner_indices = find_corner_points(coords, angle_threshold)

    # Create interpolated segments, ensuring no duplicate points at the boundaries
    segments = [
        interpolate_segment_with_original_points(coords[start:end + 1], step_size)
        for start, end in zip(corner_indices[:-1], corner_indices[1:])
    ]

    # Stack all segments but remove the last point of the last segment to avoid duplicate
    all_segments = np.vstack(segments)
    if len(all_segments) > 1:
        return all_segments[:-1]  # Remove the last point to avoid duplicate
    return all_segments  # Return the segment if there's only one


def interpolate_line(line: LineString, distances: np.ndarray) -> np.ndarray:
    coords = np.array(line.coords)

    # If closed, append first coordinate
    if line.is_closed:
        coords = np.vstack([coords, coords[0]])

    dxdy = np.diff(coords, axis=0)
    seg_lengths = np.hypot(dxdy[:, 0], dxdy[:, 1], dtype=np.float32)

    cum_lengths = np.r_[0, np.cumsum(seg_lengths)]  # Avoids explicit concatenate

    indices = np.searchsorted(cum_lengths, distances, side='right') - 1
    indices = np.clip(indices, 0, len(coords) - 2)

    seg_start = coords[indices]
    seg_end = coords[indices + 1]
    seg_dists = np.maximum(cum_lengths[indices + 1] - cum_lengths[indices], 1e-10)  # Prevent division by zero

    t = (distances - cum_lengths[indices]) / seg_dists
    t = np.clip(t, 0, 1)

    return seg_start + t[:, None] * (seg_end - seg_start)


# Local lib
import sys

from .fib import FIB
from ..beams import IonBeams


class FIBG2(FIB):

    # region Class Body

    # Class variables
    _fov_1000x = 128_000
    _fov_pixels = 65536
    _default_beam = IonBeams.Acceleration30kV.pA48

    # Instance variables
    Beams: IonBeams

    # endregion Class Body

    # region Dev Methods

    def __init__(self, magnification: int = None) -> None:
        super().__init__(magnification)
        self.Beams = IonBeams()

    def __repr__(self) -> str:
        beamlines = str(self.beam).split("\n")
        string = (f"\033[4mFEI Helios Nanolab Dualbeam G2:\033[0m\n"
                  f"\tMagnification: {self._magnification}\n"
                  f"\tField of view: {self.get_fov('um')} um\n"
                  f"\tPixelsize: {self.get_pixelsize('nm')} nm\n")
        for line in beamlines:
            string += "\t" + line + "\n"

        return string

    # endregion Dev Methods



import matplotlib.pyplot as plt
import numpy as np


def print_streamfile_time(file_path: str) -> None:
    if not file_path.endswith(".str"):
        file_path += ".str"

    pass_time = 0

    # Open the file and read all lines.
    with open(file_path, 'r') as f:
        lines = f.readlines()

    passes = int(lines[1].strip())

    # Skip the first three header lines.
    for line in lines[3:]:
        line = line.strip()
        if not line:  # Skip empty lines, if any.
            continue

        # Split the line into parts.
        parts = line.split()
        pass_time += float(parts[0])

    pass_time_us = pass_time / 10
    pass_time_s = pass_time_us * 1e-6
    total_time = pass_time_s * passes

    # Convert total seconds to HH:MM:SS format
    hours, remainder = divmod(int(total_time), 3600)
    minutes, seconds = divmod(remainder, 60)

    string = ""
    if seconds == 0:
        print("Mill time is below 0 seconds.")
        return
    else:
        if seconds > 1:
            string += str(seconds) + " seconds."
        else:
            string += str(seconds) + " second."
    if minutes != 0:
        if minutes > 1:
            string = str(minutes) + " minutes, and " + string
        else:
            string = str(minutes) + " minute, and " + string
    if hours != 0:
        if hours > 1:
            string = str(hours) + " hours, "
        else:
            string = str(hours) + " hour, "

    print("Mill time: " + string)


def plot_stream_file(file_path: str, title: str = None) -> None:

    if not file_path.endswith(".str"):
        file_path += ".str"

    segments = []  # List to store segments (each segment is a list of (x, y) tuples)
    current_segment = []  # List to store the current segment

    # Open the file and read all lines.
    with open(file_path, 'r') as f:
        lines = f.readlines()

    # Skip the first three header lines.
    for line in lines[3:]:
        line = line.strip()
        if not line:  # Skip empty lines, if any.
            continue

        # Split the line into parts.
        parts = line.split()

        # We expect at least three numbers per line.
        if len(parts) < 3:
            continue

        # The second and third elements are x and y pixel coordinates.
        # (Remember: list indices start at 0, so parts[1] is the second element.)
        try:
            x = int(parts[1])
            y = int(parts[2])
        except ValueError:
            raise ValueError(f"x and y-coordinates must be integers representing a pixel. "
                             f"Got: ({parts[1]}, {parts[2]}).")

        # Check if there is a fourth element.
        if len(parts) >= 4:
            if current_segment:
                segments.append(current_segment.copy())
            current_segment = [(x, y)]

        else:
            current_segment.append((x, y))

    segments.append(current_segment.copy())

    # Now, plot each segment.
    for seg in segments:
        # Unzip the list of (x,y) tuples into separate x and y lists.
        seg = np.array(seg)
        seg *= np.array([1, -1])
        seg += np.array([0, 65536])
        xs, ys = zip(*seg)
        plt.plot(xs, ys, "o-")  # You can remove or adjust marker style if desired.

    plt.xlabel("X Pixels")
    plt.ylabel("Y Pixels")
    plt.title("Streamlines")
    plt.xlim(0, 65535)
    plt.ylim(0, 65535)
    plt.gca().set_aspect("equal")
    plt.show()



from typing import List, Union
import itertools

import numpy as np
from numpy.typing import NDArray
from shapely import Point, LineString, MultiLineString, Polygon, MultiPolygon, buffer, prepare, intersection

from .helper_functions import interpolate_line
from ..beams import BeamType
from ..mill import Mill


def _raster_sequential(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                       horizontal: bool = True) -> List[NDArray]:

    beam_radius = beam.get_radius(units="nm")
    overlap = mill.get_overlap()
    step = 2 * beam_radius * (1 - overlap)
    dwelltime = int(mill.get_dwell_time("ns")) / 100

    # Ensure we have a MultiPolygon to loop through
    if isinstance(shape, Polygon):
        shape = MultiPolygon([shape])

    streamlines_per_polygon = []  # Store scan lines per polygon

    for polygon in shape.geoms:  # Process each polygon separately
        buffered: Polygon = buffer(polygon, -beam_radius, join_style="mitre", quad_segs=32)
        if buffered.is_empty:
            continue

        prepare(buffered)
        streamlines = []
        minx, miny, maxx, maxy = buffered.bounds

        buffered: Polygon = buffer(buffered, + beam_radius * 0.1, join_style="mitre", quad_segs=32)
        prepare(buffered)

        # Fetch some scan direction dependent values
        if horizontal:
            nr_steps = np.rint((maxy - miny) / step)
            span = (maxy - miny)
            scan_positions = np.linspace(miny, maxy, int(nr_steps))
        else:
            nr_steps = np.rint((maxx - minx) / step)
            span = (maxx - minx)
            scan_positions = np.linspace(minx, maxx, int(nr_steps))

        # Find the stepsize in the direction normal to the scan direction.
        normal_stepsize = span / nr_steps

        for current in scan_positions:
            if horizontal:
                scan_line = LineString([(minx - 1000, current), (maxx + 1000, current)])
            else:
                scan_line = LineString([(current, miny - 1000), (current, maxy + 1000)])

            inters = intersection(buffered, scan_line)

            if inters.is_empty:
                current += normal_stepsize
                continue

            segments = []
            if isinstance(inters, LineString):
                segments = [inters]
            elif isinstance(inters, MultiLineString):
                segments = list(inters.geoms)
            elif isinstance(inters, Point):
                segments = [inters]

            for seg in segments:

                if isinstance(seg, Point):
                    points = np.array(seg.coords)
                    codes = np.full(len(points), 2, dtype=int)
                    codes[0] = 1  # Start code
                    dwelltimes = np.full(len(points), dwelltime)
                    streamline = np.column_stack((dwelltimes, points[:, 0], points[:, 1], codes))
                    streamlines.append(streamline)
                    continue

                coords = np.array(seg.coords)
                if coords.size == 0:
                    continue
                sorted_indices = np.argsort(coords[:, 0])
                sorted_coords = coords[sorted_indices]
                sorted_seg = LineString(sorted_coords)

                line_length = sorted_seg.length
                if line_length == 0:
                    continue
                nr_steps = int(np.rint(line_length / step))
                distances = np.linspace(0, line_length, nr_steps)
                points = np.array([sorted_coords[0]]) if nr_steps == 0 else interpolate_line(sorted_seg, distances)

                if points.size == 0:
                    continue

                codes = np.full(len(points), 2, dtype=int)
                codes[0] = 1  # Start code
                dwelltimes = np.full(len(points), dwelltime)
                streamline = np.column_stack((dwelltimes, points[:, 0], points[:, 1], codes))
                streamlines.append(streamline)

            current += normal_stepsize

        streamlines_per_polygon.extend(streamlines)

    return streamlines_per_polygon


def horizontal_left_to_right(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                             horizontal: bool = True) -> List[NDArray]:
    streamlines = _raster_sequential(shape, mill, beam, horizontal=True)
    tolerance = beam.get_radius("nm")
    streamlines.sort(key=lambda s: (int(s[0, 2] / tolerance), s[0, 1]))
    return streamlines


def horizontal_right_to_left(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                             horizontal: bool = True) -> List[NDArray]:

    streamlines = _raster_sequential(shape, mill, beam, horizontal=True)

    tolerance = beam.get_radius("nm")
    streamlines.sort(key=lambda s: (int(s[0, 2] / tolerance), s[0, 1]))

    grouped_streamlines = []
    for group_key, group in itertools.groupby(streamlines, key=lambda s: int(s[0, 2] / tolerance)):
        grouped_streamlines.append(list(group))

    # Now, for every other pass (group), reverse the order of points in each scan line.
    final_streamlines = []
    for i, group in enumerate(grouped_streamlines):
        group = [line[::-1] for line in group][::-1]
        final_streamlines.extend(group)

    return final_streamlines


def vertical_bottom_to_top(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                           horizontal: bool = True) -> List[NDArray]:
    streamlines = _raster_sequential(shape, mill, beam, horizontal=False)
    tolerance = beam.get_radius("nm")
    streamlines.sort(key=lambda s: (int(s[0, 1] / tolerance), s[0, 2]))
    return streamlines


def vertical_top_to_bottom(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                           horizontal: bool = True) -> List[NDArray]:

    streamlines = _raster_sequential(shape, mill, beam, horizontal=False)

    tolerance = beam.get_radius("nm")
    streamlines.sort(key=lambda s: (int(s[0, 1] / tolerance), s[0, 2]))

    # Group the streamlines by the bucketed y coordinate.
    grouped_streamlines = []
    for group_key, group in itertools.groupby(streamlines, key=lambda s: int(s[0, 1] / tolerance)):
        grouped_streamlines.append(list(group))

    # Now, for every other pass (group), reverse the order of points in each scan line.
    final_streamlines = []
    for i, group in enumerate(grouped_streamlines):
        group = [line[::-1] for line in group][::-1]
        final_streamlines.extend(group)

    return final_streamlines



# Std lib
from typing import List, Union

# 3rd party lib
import numpy as np
from numpy.typing import NDArray
from shapely import buffer, Polygon, MultiPolygon, prepare

# Local lib
from .helper_functions import interpolate_curve
from ..beams import BeamType
from ..mill import Mill


def contour_inward(shape: Union[Polygon, MultiPolygon], mill: Mill, beam: BeamType,
                   rasterize: bool = True) -> List[NDArray]:
    beam_radius = beam.get_radius(units="nm")
    overlap = mill.get_overlap()
    step = 2 * beam_radius * (1 - overlap)
    dwelltime = int(mill.get_dwell_time("ns")) / 100

    contours = []
    prepare(shape)  # Precompute spatial indexes for faster operations

    def add_linestrings(geom):
        if isinstance(geom, Polygon):
            if not geom.exterior.is_empty:
                contours.append(geom.exterior)
            contours.extend(geom.interiors)
        elif isinstance(geom, MultiPolygon):
            for poly in geom.geoms:
                if not poly.exterior.is_empty:
                    contours.append(poly.exterior)
                contours.extend(poly.interiors)

    buffered = buffer(shape, -beam_radius, join_style="mitre")
    quad_segs = 32  # Reduced from 100 for initial buffer
    k = 0

    while True:
        if k != 0:
            quad_segs = max(8, int(quad_segs * 0.8))  # Decrease resolution inward
            buffered = buffer(buffered, -step, join_style="round", quad_segs=quad_segs)
            if buffered.is_empty:
                break
        add_linestrings(buffered)
        k += 1

    streamlines = []
    if not rasterize:
        for line in contours:
            coords = np.array(line.coords)
            streamline = np.column_stack((
                np.zeros(len(coords)),
                coords[:, 0],
                coords[:, 1],
                np.zeros(len(coords))
            ))
            streamlines.append(streamline)
        return streamlines

    for line in contours:
        line_length = line.length
        nr_steps = int(np.rint(line_length / step))

        if line.is_closed and nr_steps == 3:
            if line_length / 2 > beam_radius * 1.5:
                points = interpolate_curve(line, step)
                centroid = line.centroid
                if line.distance(centroid) > beam_radius:
                    points = np.vstack([points, [centroid.x, centroid.y]])
            else:
                points = np.array([[line.centroid.x, line.centroid.y]])
        elif not line.is_closed and nr_steps == 2:
            points = np.array([[line.centroid.x, line.centroid.y]])
        else:
            points = interpolate_curve(line, step)

        if points.size == 0:
            continue

        codes = np.full(len(points), 2, dtype=int)
        codes[0] = 1
        dwelltimes = np.full(len(points), dwelltime)
        streamline = np.column_stack((dwelltimes, points[:, 0], points[:, 1], codes))

        streamlines.append(streamline)

    return streamlines


def contour_outward(shape: Polygon | MultiPolygon, mill: Mill, beam: BeamType, rasterize: bool = True) -> List[NDArray]:
    return contour_inward(shape, mill, beam, rasterize)[::-1]

